import pickle
from IPython.utils import io
from datetime import timedelta
from os.path import exists
from ib_insync import Option
from ibtools.tools import getApplication, toTWSDateFromDate, today, toDate, toDates, OptionDetail


class OptionChain:

    def __init__(self, underlying, expiration, calls, puts):
        self.underlying = underlying
        self.symbol = underlying.symbol
        self.expiration = expiration
        self.calls = calls
        self.puts = puts
        self.callContracts = self.__getContracts(self.calls)
        self.putContracts = self.__getContracts(self.puts)

    def __getContracts(self, detailsByStrike):
        return [detail.option for detail in detailsByStrike.values()]

    def __str__(self):
        return 'Option chain for '+str(self.symbol) + ' on '+str(self.expiration)


def getOptionChains(underlyingContract, *dates):
    getApplication().qualifyContracts(underlyingContract)
    expirations = _filterExpirations(toDates(*dates), underlyingContract)
    return _chainsForExpirations(underlyingContract, expirations)


def getOptionChainsInDateRange(underlyingContract, beginDate, endDate):
    getApplication().qualifyContracts(underlyingContract)
    chainExpirations = _expirationsOfChain(underlyingContract)
    expirationsInRange = _expirationsInDateRange(toDate(beginDate),
                                                 toDate(endDate),
                                                 chainExpirations)
    return _chainsForExpirations(underlyingContract, expirationsInRange)


def getOptionChainsInDteRange(underlyingContract, lowerDte, higherDte):
    currentDay = today()
    return getOptionChainsInDateRange(underlyingContract,
                                      currentDay+timedelta(days=lowerDte),
                                      currentDay+timedelta(days=higherDte))


###################################################################


def _filterExpirations(dates, contract):
    chainExpirations = _expirationsOfChain(contract)
    return [date for date in dates if date in chainExpirations]


def _chainsForExpirations(underlyingContract, expirations):
    storedChains = _loadValidChains(underlyingContract)
    print(underlyingContract.symbol+":storedChains "+str(storedChains))
    cachedChains = _cachedChains(storedChains, expirations)
    print(underlyingContract.symbol+":cachedChains "+str(cachedChains))
    newChains = _newChains(underlyingContract, storedChains, expirations)
    print(underlyingContract.symbol+":newChains "+str(newChains))

    return _serializeAndReturnChains(underlyingContract, cachedChains, newChains, storedChains)


def _cachedChains(storedChains, expirations):
    storedExpirations = list(storedChains.keys())
    cachedExpirations = _cachedExpirations(storedExpirations, expirations)

    return {expiration: storedChains[expiration] for expiration in cachedExpirations}


def _newChains(underlyingContract, storedChains, expirations):
    storedExpirations = list(storedChains.keys())
    notCachedExpirations = _notCachedExpirations(
        storedExpirations, expirations)

    return {expiration: _createContractsForExpiration(underlyingContract, expiration)
            for expiration in notCachedExpirations}


def _serializeAndReturnChains(underlyingContract, cachedChains, newChains, storedChains):
    chainsToReturn = {**cachedChains, **newChains}
    chainsToSerialize = {**storedChains, **newChains}
    _serializeChains(underlyingContract, chainsToSerialize)
    return chainsToReturn


def _expirationsInDateRange(beginDate, endDate, expirations):
    return [expiration for expiration in expirations if beginDate <= expiration <= endDate]


def _cachedExpirations(storedExpirations, requestedExpirations):
    return [expiration for expiration in requestedExpirations if expiration in storedExpirations]


def _notCachedExpirations(storedExpirations, requestedExpirations):
    return [expiration for expiration in requestedExpirations if not expiration in storedExpirations]


def _chain(contract):
    optChain = getApplication().reqSecDefOptParams(underlyingSymbol=contract.symbol,
                                                   futFopExchange='',
                                                   underlyingSecType='STK',
                                                   underlyingConId=contract.conId)
    return _filterExchangeFromChain(optChain, 'SMART')


def _optionDetailsForExpiration(underlying, chain, right, expiration):
    options = [Option(chain.tradingClass,
                      toTWSDateFromDate(expiration),
                      strike,
                      right,
                      chain.exchange)
               for strike in chain.strikes]
    with io.capture_output():  # Trying to suppress Error 200 from TWS
        validOptions = getApplication().qualifyContracts(*options)
    return [OptionDetail(option, underlying) for option in validOptions]


def _contractsByStrikes(underlying, chain, right, expiration):
    optionDetails = _optionDetailsForExpiration(
        underlying, chain, right, expiration)
    return {optionDetail.strike:  optionDetail for optionDetail in optionDetails}


def _createContractsForExpiration(underlying, expiration):
    print("Creating option contracts for " +
          underlying.symbol+" "+str(expiration)+" ...")

    chain = _chain(underlying)
    callContracts = _contractsByStrikes(underlying, chain, "C", expiration)
    putContracts = _contractsByStrikes(underlying, chain, "P", expiration)

    print("Created option contracts for " +
          underlying.symbol+" "+str(expiration)+".")
    return OptionChain(underlying, expiration, callContracts, putContracts)


def _expirationsOfChain(contract):
    return toDates(*_chain(contract).expirations)


def _timeDeltaInDays(earlyDate, lateDate):
    return (lateDate - earlyDate).days


def _dte(expiration):
    return _timeDeltaInDays(today(), expiration)


def _filterExchangeFromChain(chain, exchange):
    return next(filter(lambda x: x.exchange == exchange, chain))


def _fileNameForChains(contract):
    return contract.symbol+'_'+str(contract.conId)+'_optionchains'


def _deSerializeChains(contract):
    filename = _fileNameForChains(contract)
    if exists(filename) == False:
        emptyChains = dict()
        _serializeChains(contract, emptyChains)
        return emptyChains

    infile = open(filename, 'rb')
    storedChains = pickle.load(infile)
    infile.close()
    return storedChains


def _serializeChains(contract, chains):
    filename = _fileNameForChains(contract)
    outfile = open(filename, 'wb')
    pickle.dump(chains, outfile)
    outfile.close()


def _filterOutdatedChains(storedChains):
    return {expiration: chain for expiration, chain in storedChains.items()
            if _dte(expiration) >= 0}


def _loadValidChains(contract):
    storedChains = _deSerializeChains(contract)
    validStoredChains = _filterOutdatedChains(storedChains)
    _serializeChains(contract, validStoredChains)
    return validStoredChains



import datetime
from rx.subject import Subject


class OptionDetail:

    def __init__(self, option, underlying):
        _app.qualifyContracts(underlying)

        self.option = option
        self.underlying = underlying
        self.symbol = underlying.symbol
        self.expiration = toDate(option.lastTradeDateOrContractMonth)
        self.strike = option.strike
        self.right = option.right

    def __str__(self):
        return 'OptionDetail:\n' + \
            'Option: '+str(self.option) + '\n' + \
            'Underlying: ' + str(self.underlying)


def setApplication(app):
    global _app
    _app = app
    global _marketDataObservable
    _marketDataObservable = _MarketDataStream()


def getApplication():
    return _app


def toDate(date):
    return toDateFromTWSDate(date)


def toDates(*dates):
    return [toDate(date) for date in dates]


def today():
    return datetime.date.today()


def toDateFromTWSDate(twsDate):
    if isinstance(twsDate, str):
        return datetime.datetime.strptime(twsDate, _twsDateFormat).date()
    return twsDate


def toTWSDateFromDate(date):
    if isinstance(date, str):
        return date
    return date.strftime(_twsDateFormat)


###################################################################
_app = None
_marketDataObservable = None
_twsDateFormat = '%Y%m%d'


class _MarketDataStream:
    def __init__(self):
        self.observable = Subject()
        _app.pendingTickersEvent += self.__onMarketData

    def __onMarketData(self, pendingTickers):
        for marketData in pendingTickers:
            self.observable.on_next(marketData)


def _marketDataObservable():
    return _marketDataObservable




import math
import rx
from rx.subject import Subject
from rx.operators import *
from IPython.utils import io
from ibtools.tools import getApplication, _marketDataObservable


class OptionMarketData:

    def __init__(self, optionDetail):
        self.optionDetail = optionDetail
        self.symbol = optionDetail.symbol
        self.expiration = optionDetail.expiration
        self.strike = optionDetail.strike
        self.right = optionDetail.right
        self.isDataAvailable = False

    def subscribe(self, marketDataReadyListener=lambda snapshot: None):
        self.__marketDataReadyListener = marketDataReadyListener
        self.undMarketData = _requestMarketData(self.optionDetail.underlying)

        _marketDataObservable().observable \
            .pipe(
                filter(lambda data: data.contract == self.optionDetail.option),
                filter(_isMarketDataReady),
                distinct(lambda data: data.contract),
                take(1)
        ) .subscribe(on_completed=self.__onMarketDataComplete)

        self.marketData = _requestMarketData(self.optionDetail.option)

    def __onMarketDataComplete(self):
        self.isDataAvailable = True
        self.__marketDataReadyListener(self)
        del self.__marketDataReadyListener

    def unsubscribe(self):
        if hasattr(self, 'undMarketData'):
            with io.capture_output():
                _cancelMarketData(self.optionDetail.underlying)
                _cancelMarketData(self.optionDetail.option)
            del self.undMarketData
            del self.marketData
            self.isDataAvailable = False

    def __nonzero__(self):
        return self.isDataAvailable

    def __str__(self):
        return 'Option market data '+str(self.symbol) + \
            ' on ' + str(self.expiration) + \
            ' at strike ' + str(self.strike) + str(self.right)


class OptionChainMarketData:

    def __init__(self, optionChain):
        self.optionChain = optionChain
        self.underlying = optionChain.underlying
        self.symbol = optionChain.symbol
        self.expiration = optionChain.expiration
        self.calls = _fromOptionDetailsToOptionMarketData(optionChain.calls)
        self.puts = _fromOptionDetailsToOptionMarketData(optionChain.puts)
        self.__allOptionMarketData = list(self.calls.values()) + \
            list(self.puts.values())
        self.__chainSize = len(self.calls) + len(self.puts)
        self.__isDataAvailable = False

    def subscribe(self, marketDataReadyListener=lambda snapshot: None):
        self.__marketDataReadyListener = marketDataReadyListener
        _subscribeAllItems(self.__chainSize,
                           self.__onAllSnapshotsSubscribed,
                           self.__subscibeAllOptions)

    def __subscibeAllOptions(self, observable):
        for optionMarketData in self.__allOptionMarketData:
            optionMarketData.subscribe(observable.on_next)
            getApplication().sleep(_markteReqThrottle)

    def __onAllSnapshotsSubscribed(self):
        self.__isDataAvailable = True
        print('Option chain market data for ' + self.symbol +
              ' on '+str(self.expiration) + ' is now subscribed.')
        self.__marketDataReadyListener(self)
        del self.__marketDataReadyListener

    def unsubscribe(self):
        for optionMarketData in self.__allOptionMarketData:
            optionMarketData.unsubscribe()
        self.__isDataAvailable = False
        print('Option chain market data for ' + self.symbol +
              ' on '+str(self.expiration) + ' is now unsubscribed.')

    def __len__(self):
        return self.__chainSize

    def __nonzero__(self):
        return self.__isDataAvailable

    def __str__(self):
        return 'Option chain market data for ' + str(self.symbol) + ' on '+str(self.expiration)


class OptionChainsMarketData:

    def __init__(self, optionChains):
        self.optionChains = optionChains
        self.symbol = next(iter(optionChains.values())).symbol
        self.expirations = list(optionChains.keys())
        self.chains = {chainMarketData.expiration: chainMarketData
                       for chainMarketData in [OptionChainMarketData(chain)
                                               for chain in optionChains.values()]}
        self.__noOfChains = len(self.chains)
        self.__isDataAvailable = False

    def subscribe(self, marketDataReadyListener=lambda snapshot: None):
        self.__marketDataReadyListener = marketDataReadyListener
        _subscribeAllItems(len(self.optionChains.values()),
                           self.__onAllChainMarketDataSubscribed,
                           self.__subscibeAllChains)

    def __subscibeAllChains(self, observable):
        for optionChainMarketData in self.chains.values():
            optionChainMarketData.subscribe(observable.on_next)

    def unsubscribe(self):
        for optionChainMarketData in self.chains.values():
            optionChainMarketData.unsubscribe()
        self.__isDataAvailable = False
        print('Option chains market data for ' + self.symbol +
              ' on expirations '+str(self.expirations) + ' are now unsubscribed.')

    def __onAllChainMarketDataSubscribed(self):
        self.__isDataAvailable = True
        print('Option chains market data for ' + self.symbol +
              ' on expirations '+str(self.expirations) + ' are now subscribed.')
        self.__marketDataReadyListener(self)
        del self.__marketDataReadyListener

    def __nonzero__(self):
        return self.__isDataAvailable

    def keys(self):
        return self.chains.keys()

    def __getitem__(self, expiration):
        return self.chains[expiration]

    def __len__(self):
        return self.__noOfChains

    def __str__(self):
        return 'Option chains market data for ' + \
            str(self.symbol) + ' on expirations ' + \
            str(self.expirations)

###################################################################


_markteReqThrottle = 0.1

_putCallVolume = 100
_openInterest = 101
_histVolatility = 104
_avOptionVolume = 105
_impliedVolatility = 106
_tickList = [_putCallVolume, _openInterest,
             _histVolatility, _avOptionVolume, _impliedVolatility]
_genericTickList = ','.join([str(i) for i in _tickList])


def _isMarketDataReady(marketData):
    # return True
    return not math.isnan(marketData.bidSize)
    # return not math.isnan(marketData.callOpenInterest) and \
    #   not math.isnan(marketData.bid) \
    #  and hasattr(marketData, 'modelGreeks')


def _requestMarketData(contract):
    return getApplication().reqMktData(contract, genericTickList=_genericTickList)


def _cancelMarketData(contract):
    getApplication().cancelMktData(contract)


def _fromOptionDetailsToOptionMarketData(optionDetails):
    return rx.of(*list(optionDetails.values())).pipe(
        map(lambda detail: OptionMarketData(detail)),
        to_dict(lambda marketData: marketData.strike,
                lambda marketData: marketData)
    ).run()


def _subscribeAllItems(noOfItems, onCompleted, subscribeAllFunc):
    observable = Subject()
    observable.pipe(
        take(noOfItems)
    ) .subscribe(on_completed=onCompleted)
    subscribeAllFunc(observable)



from .tools import *

from .optionchain import OptionChain
from .optionchain import getOptionChains
from .optionchain import getOptionChainsInDateRange
from .optionchain import getOptionChainsInDteRange

from .optionsnapshot import OptionMarketData
from .optionsnapshot import OptionChainMarketData
from .optionsnapshot import OptionChainsMarketData



from setuptools import setup, find_packages

VERSION = '0.6'
DESCRIPTION = 'Little helper tools for IB API'
LONG_DESCRIPTION = 'Little helper tools for IB API'

setup(
    name="ibtools",
    version=VERSION,
    author="Juergen Reiss",
    author_email="juxeiier@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url='https://github.com/juxeii/ibtools',
    license='BSD',
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=['rx', 'ib_insync', 'IPython'],
    keywords=['python', 'first package'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        'Topic :: Office/Business :: Financial :: Investment',
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        'License :: OSI Approved :: BSD License'
    ]
)
